//---------------------------------------------------------------------------

#include <windows.h>

#include <fstream>
#include <iostream>
#include <vector>
#include <map>
#include <string>

#include <cstring>
#include <cstdio>

#include <windows.h>

#include "ucl/file.hpp"
#include "charmap.h"

#pragma hdrstop

//---------------------------------------------------------------------------


#define FLAGS_DELETEOLD 0x01
#define FLAGS_TEMPORARY 0x02
#define FLAGS_IMPORT    0x04
char simparam[] = "owb2sim_simexit_call";

const BYTE modulesig[] = {'O','B','I','M','P','O','R','T'};
#define OWBIMPORT_VER 0x00000003ul
#define CBSIGOFFS	2032




using namespace std;


BYTE flags = FLAGS_DELETEOLD;
string projpath;
BYTE* owbdata;
string exepath;
string simpath;
DWORD waittime;

map <string, string> filenames;


string getExePath (void);
void syntaxMessage (void);
int evaluateParameters (int argc, char* argv[]);
int writeBinFile (void);
HWND findSIMWindow (void);
bool startProcess (string path, PROCESS_INFORMATION& pi);
void sendKeyboardEvents (void);
void loadSettings (void);
int writeRegistryKey (void);
int printErrorMessage (void);




string getExePath (void)
{
	char exepath[1024];
	int i;

	GetModuleFileName (NULL, exepath, 1024);
	
	return (ucl::extractFilePath (exepath));
}

void syntaxMessage (void)
{
	cout << endl << endl
		 << "Syntax: owb2sim [-t|-T][-e] <makefile>" << endl << endl
		 << " -t: copy files and erase old ones (default)" << endl
		 << " -T: copy files" << endl		
		 << " -e: temporary mode (changes in virtual PV are not saved)" << endl << endl;
}

int evaluateParameters (int argc, char* argv[])
{
	if (argc == 1) // no arguments given
	{
		syntaxMessage ();
		return (1);
	}


		/* evaluate parameters:
		 *
		 * OWB2SIM [-t|-T][-e] <makefile>
		 *
		 * -t: copy files and erase old ones (default)
		 * -T: copy files
		 * -e: temporary mode (changes in virtual PV are not saved)
		 *
		 */
	for (int i = 1; i < argc; ++i)
	{
		if ((argv[i][2] == '\0') && ((argv[i][0] == '-') || (argv[i][0] == '/')))
			switch (argv[i][1])
			{
			case 'e':
				flags |= FLAGS_TEMPORARY;
				continue;
			case 'T':
				flags &= ~FLAGS_DELETEOLD;
			case 't':
				continue;
			}
			
		if (!projpath.empty ())
		{
			syntaxMessage ();
			cerr << "Error: Too many parameters" << endl;
			return (1);
		}
		projpath = argv[i];
		continue;
	}
		 
	if (projpath.empty ())
	{
		syntaxMessage ();
		cerr << "Error: No file name given" << endl;
		return (1);
	}
	return (0);
}


template <typename T> void xchg (T& t1, T& t2)
{
	T t3 = t1;
	t1 = t2;
	t2 = t3;
}

enum sfbkpmode_t {sfBackup, sfRestore, sfDelete};
int backupSIMFiles (sfbkpmode_t mode)
{
	static const char* files[] = {
		"\\SIM\\OWBSRAM.BIN",
		/*"\\BIOS\\OWBBIOS.00",*/
		/*"\\C\\BIN\\APLOBIMP.BIN",*/
		"\\SIM\\OBFLASH32.BIN"
	};
	const FILENUM = 2 /* 4 */;
	string bkppath;
	const char* p1;
	const char* p2;
	

	for (int n = 0; n < FILENUM; ++n)
	{
		p1 = files[n];
		bkppath = ucl::extractFilePath (p1) + ucl::extractFileName (p1) + ".TMP";
		p2 = bkppath.c_str ();

        switch (mode)
        {
        case sfRestore:
			xchg (p1, p2);
        case sfBackup:
            if (!CopyFile ((simpath + p1).c_str (), (simpath + p2).c_str (), FALSE))
            {
                cerr << "Error: Unable to create backup of " << simpath << p1 << endl;
                return (1);
            }
            break;
        case sfDelete:
			DeleteFile ((simpath + p2).c_str ());
            break;
        }
	}

	return (0);
}


int writeFileSection (const char* workdir, int cat, WORD*& filecnt, 
					  DWORD*& fileoffs, BYTE* data, BYTE*& bptr, BYTE*& bptr_beg,
					  const ucl::IniFile::Section* files, ucl::IniFile& accessfile)
{
	FILE* 		file;
	int 		i, l;
	HANDLE		hFile;
	FILETIME 	ftLastWriteTime;
	char		buf[300];
	char 		c;
	string		sbuf;
    char*       bufptr;
    bool        unchanged = false;

	l = files->lineCount ();
	for (i = 0; i < l; ++i)
	{
		if ((*files)[i].empty ())
			continue;
		hFile = CreateFile ((*files)[i].c_str (), GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE)
		{
			cerr << "Error: File \"" << (*files)[i] << "\" not found" << endl;
			return (1);
		}
		GetFileTime (hFile, NULL, NULL, &ftLastWriteTime);
		CloseHandle (hFile);


		file = fopen ((*files)[i].c_str (), "r");
		if (!file)
		{
			cerr << "Error: File \"" << (*files)[i] << "\" not found" << endl;
			return (1);
		}
		fseek (file, 0, SEEK_SET);
		
			/* copy title to buf */
		bufptr = buf;
		do
			*(bufptr++) = (c = fgetc (file));
		while ((c != '\n') && (c != EOF));
		bufptr[-1] = '\0';

		if (strlen (buf) < 3)
		{
			cerr << (*files)[i] << ",1: no file name given" << endl;
			return (1);
		}
		strlwr (buf + 2);
		sbuf = (ucl::extractFilePath ((*files)[i]).empty () ? workdir : "") + (*files)[i];
		
		filenames[string (buf + 2)] = sbuf;
		
			/* if file was not changed, do not upload */
		if (!(flags & FLAGS_TEMPORARY))
			unchanged = ((accessfile["LastWriteTimeLow"][(*files)[i]].asInteger ()
					== static_cast <long> (ftLastWriteTime.dwLowDateTime))
			 && (accessfile["LastWriteTimeHigh"][(*files)[i]].asInteger ()
					== static_cast <long> (ftLastWriteTime.dwHighDateTime)));
			/*{
				fclose (file);
				continue;
			}*/

			/* update last write time */
		if (!(flags & FLAGS_TEMPORARY) && !unchanged)
		{
			itoa (ftLastWriteTime.dwLowDateTime, buf, 10);
			accessfile["LastWriteTimeLow"][(*files)[i]] = buf;
			itoa (ftLastWriteTime.dwHighDateTime, buf, 10);
			accessfile["LastWriteTimeHigh"][(*files)[i]] = buf;
		}
		
		fseek (file, 0, SEEK_SET);
		*(fileoffs++) = bptr - (owbdata + 0x20000);
		++*filecnt;
		*(bptr++) = static_cast <char> (cat);
		*(bptr++) = static_cast <char> (unchanged);
		do
		{
			c = fgetc (file);
			if (c == EOF)
				break;
			*(bptr++) = charmap_ansi2pv[static_cast <unsigned char> (c)];
            if (bptr > (data + 0x40000))
            {
                cerr << sbuf << ",1: Too many files" << endl;
                return (1);
            }
		} while (true);
		*(bptr++) = '\0';
		fclose (file);
		if ((bptr - bptr_beg - 2) > 0x0800)
		{
			cerr << sbuf << ",1: File size exceeds 2048 bytes (size: "
				 << (bptr - bptr_beg - 2) << " bytes)" << endl;
			return (1);
		}
		bptr_beg = bptr;
	}
	return (0);
}
int writeBinFile (void)
{
	ucl::IniFile accessfile (ucl::extractFilePath (projpath)
						   + ucl::extractFileName (projpath) + ".obc");

	FILE* 		file;
	BYTE* 		bptr;
	BYTE*		bptr_beg;
	WORD* 		filecnt;
	DWORD*  	fileoffs;
	int 		i;
	ucl::IniFile projfile (projpath);
	char		workdir[512];
	string		sbuf;
	int			cat, version;


	version = projfile["Settings"]["ProjectFormatVersion"].asInteger ();
	cat = projfile["Settings"]["Category"].asInteger ();
	
	GetCurrentDirectory (512, workdir);
	strcat (workdir, "\\");

	owbdata = new BYTE[0x40000]; /* 256 kB */

		/* read OWBasic binary */
	file = fopen ((exepath + "OWBSIM.dat").c_str (), "rb");
	if (!file)
	{
		cerr << "Error: OWBSIM.dat not found" << endl;
		return (1);
	}
	fread (owbdata, 1, 0x20000, file);
	fclose (file);

	
		/* append file transfer data */
	bptr = owbdata + 0x20000;
	
	memcpy (bptr, modulesig, 8); /* signature */
	bptr += 8;
	
	*(reinterpret_cast <DWORD*> (bptr)) = OWBIMPORT_VER; /* version */
	bptr += sizeof (DWORD);
	
	filecnt = reinterpret_cast <WORD*> (bptr); *filecnt = 0; /* file count */
	bptr += sizeof (WORD);
	
	*(bptr++) = cat;
	*(bptr++) = flags;

	strcpy (bptr, projfile["Settings"]["Progname"].asString ().c_str ());
	bptr += 32;

	fileoffs = reinterpret_cast <DWORD*> (bptr);
	bptr += sizeof (DWORD) * 64;


	bptr_beg = bptr;

	if (version >= 120)
		for (i = 1; i <= 5; ++i)
		{
			sbuf = string ("Category ") + static_cast <char> ('0' + i);
			if (!projfile.sectionExists (sbuf))
				continue;
			if (writeFileSection (workdir, i, filecnt, fileoffs, owbdata, bptr,
								   bptr_beg, &projfile[sbuf], accessfile))
				return (1);
		}
	else
		if (writeFileSection (workdir, cat, filecnt, fileoffs, owbdata, bptr,
							  bptr_beg, &projfile["Files"], accessfile))
			return (1);

	if (!(flags & FLAGS_TEMPORARY))
		accessfile.saveToFile (ucl::extractFilePath (projpath) + ucl::extractFileName (projpath) + ".obc");

	file = fopen ((simpath + "C\\BIN\\OWB4OIMP.BIN").c_str (), "wb");
	if (!file)
	{
		cerr << "Error: Unable to write output file" << endl;
		return (1);
    }
	fseek (file, 0, SEEK_SET);
	fwrite (owbdata, 1, bptr - owbdata, file);
	fclose (file);

	return (0);
}

HWND findSIMWindow (void)
{
	HWND hSIMWin;
	char buf[512];
    string simtitle = "SIM3022 - " + simpath + "SIM\\";

    simtitle += (flags & FLAGS_TEMPORARY) ? "obimptmp.cpj" : "obimport.cpj";
	hSIMWin = FindWindow ("TMainForm", simtitle.c_str ());
	if (!hSIMWin)
	{
		hSIMWin = FindWindow ("TMainForm", NULL);
		if (!hSIMWin)
		{
			cerr << "Error: SIM window not found" << endl;
			return (NULL);
		}
		
		GetWindowText (hSIMWin, buf, 512);
		buf[7] = '\0';
		if (strcmp (buf, "SIM3022"))
		{
			cerr << "Error: SIM window not found" << endl;
			return (NULL);
		}
	}

	SetWindowPos (hSIMWin, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	return (hSIMWin);
}

bool startProcess (string path, PROCESS_INFORMATION& pi)
{
	STARTUPINFO si;
	

	si.cb = sizeof (STARTUPINFO);
	si.lpReserved = NULL;
	si.lpDesktop = NULL;
	si.lpTitle = NULL;
	si.cbReserved2 = 0;
	si.lpReserved2 = NULL;
	si.dwFlags = 0;
	return (CreateProcess (NULL, &path[0], NULL, NULL, FALSE,
		NORMAL_PRIORITY_CLASS, NULL, ucl::extractFilePath (path).c_str (), &si, &pi));
}

void sendKeyboardEvents (void)
{

	keybd_event (VK_MENU, 0, 0, 0); /* Alt+F */
	keybd_event ('F', 0, 0, 0);
	keybd_event ('F', 0, KEYEVENTF_KEYUP, 0);
	keybd_event (VK_MENU, 0, KEYEVENTF_KEYUP, 0);

	keybd_event ('R', 0, 0, 0); /* R */
	keybd_event ('R', 0, KEYEVENTF_KEYUP, 0);
	
	keybd_event ('4', 0, 0, 0); /* 4 */
	keybd_event ('4', 0, KEYEVENTF_KEYUP, 0);
	
	Sleep (waittime);

	keybd_event (VK_F9, 0, 0, 0); /* F9 */
	keybd_event (VK_F9, 0, KEYEVENTF_KEYUP, 0);
}


void loadSettings (void)
{
	ucl::IniFile settings (exepath + "settings.ini");
	simpath = settings["Settings"]["SIMPath"];
	waittime = settings["Settings"]["DelayTime"].asInteger ();	
}

int writeRegistryKey (void)
{
	HKEY hKey = NULL;
	string regcmd = simpath + "SIM\\obimport.cpj";

    if (flags & FLAGS_TEMPORARY)
        regcmd = simpath + "SIM\\obimptmp.cpj";

	
		/* write project entry to registry */
	if (RegCreateKey (HKEY_CURRENT_USER, "Software\\CASIO\\SIM3022\\RecentFiles",
					  &hKey) != ERROR_SUCCESS)
	{
		cerr << "Error: Unable to write registry key" << endl;
		return (1);
	}
	RegSetValueEx (hKey, "4", 0, REG_SZ, regcmd.c_str (), regcmd.size () + 1);
	RegCloseKey (hKey);
	return (0);
}

int printErrorMessage (void)
{
	FILE* file;
	char buf[300 + 1];
	char* bufptr = buf;
	char c;
    string rfname = simpath + "\\SIM\\OWBSRAM";

    rfname += (flags & FLAGS_TEMPORARY) ? ".TMP" : ".BIN";

		/* readout error message from PV clipboard
			and return it on command line */
	file = fopen (rfname.c_str (), "rb");
	if (!file)
	{
		cerr << "Error: Unable to load OWBSRAM"
             << ucl::extractFileExt (rfname) << endl;
		return (1);	
	}
	fseek (file, 0x184C0 + CBSIGOFFS, SEEK_SET);
	fread (buf, 12, 1, file);
	if ((*((DWORD*) (buf + 8)) != OWBIMPORT_VER) || memcmp (buf, modulesig, 8))
    	return (0);
	
	fseek (file, 0x184C0, SEEK_SET);
	do
	{
		c = fgetc (file);
		*(bufptr++) = (c == 0x0D) ? ' ' : c;
	} while (c != 0x00);
	fclose (file);
	
	bufptr = buf;
	if (buf[0] != '\0') /* error occured */
	{
		while (*bufptr != ',')
			++bufptr;
		*(bufptr++) = '\0';
		strlwr (buf);
		cerr << filenames[buf] << ',' << bufptr << endl;
		return (1);
	}
	return (0);
}


int main (int argc, char* argv[])
{
	PROCESS_INFORMATION pi, pi2;


	exepath = getExePath ();
	loadSettings ();
	
	cout << "OWB2SIM v1.20 (c) 2006 by AUDACIA Software" << endl;


	try
	{
		if (evaluateParameters (argc, argv))
			return (1);

		if (flags & FLAGS_TEMPORARY)
			if (backupSIMFiles (sfBackup))
				return (1);

		if (writeBinFile ())
			return (1);
	
		if (writeRegistryKey ())
			return (1);
	}
	catch (char*)
	{
		cerr << "Error: Section entry not found" << endl;
		return (1);
	}
	catch (...)
	{
		cerr << "Error: Unknown error" << endl;
		return (1);
    }

		/* startup SIM */
	startProcess (simpath + "SIM\\Sim3022.exe", pi);
	
	Sleep (waittime + 100);
	
	if (!findSIMWindow ())
		return (1);

	sendKeyboardEvents ();

		/* startup SIMExit */
	startProcess (exepath + "SIMExit.exe " + simparam, pi2);
	WaitForSingleObject (pi.hProcess, INFINITE);


	if (printErrorMessage ())
		return (1);

	if (flags & FLAGS_TEMPORARY)
		if (backupSIMFiles (sfDelete))
			return (1);

	return (0);
}
//---------------------------------------------------------------------------
