/*
 * AddInInfo
 *
 * Author: J. Steingraeber
 * Date: 2001/02/27
 *
 * Changes and optimizations by Moritz Beutel, 2006/06/05
 *
 * Released to the public domain
 */

#include <stdrom.h>
#include "define.h"
#include "libc.h"
#include "l_define.h"
#include "l_libc.h"

#include "AddInInfo.h"

void far startaddin (unsigned char StartSegHighByte,
                    unsigned int FirstBlock, int n);

#if FALSE
int far deladdin (unsigned int wSeg, unsigned int wOffs);
#endif

int  getmodel (void);     /* identify PV model */

static unsigned int far modeladdr[] = {0x0000, 0x6ABA, 0x6682, 0x6682, 0x65E6};


void StartAddIn (unsigned long lLinAddr)
{
  unsigned int far* p = MK_FP (0x0000, 0x0000);
  unsigned int wSeg, wOffs;
  unsigned long far* pDW;
  unsigned int lLength;
  unsigned int model;
  /*unsigned int lin = lLinAddr;*/
  model = getmodel ();
  SwitchBank (((lLinAddr >> 16) & 0xFFFE), 1);

  wSeg = 0x2000;
  wOffs = lLinAddr & 0xFFFF;
  if (lLinAddr & 0x00010000)
    wSeg = 0x3000;

  pDW = MK_FP (wSeg, wOffs);
  lLength = (pDW[9] + 0x3FFF) >> 14; /* get length from Add-In header */

  lLinAddr = (lLinAddr >> 14) & 0xFFFF;

  p = MK_FP (0x0000, modeladdr[model]);
  if (p)
  {
    p[0] = lLinAddr;
    p[1] = lLength;
  }
  else
    LibPutMsgDlg ("WARNING!\r\r"
                  "Unrecognized PVOS!\r"
                  "Use on your own risk!\r"
                  "Better switch off now!");

  startaddin (0x80, lLinAddr | 0x8000, (word) lLength);
}

#if FALSE
int CanDelAddIn (int iNr)
{
  unsigned int far* p;
  unsigned long LinAddr;
  unsigned int model;


  model = getmodel ();

      /* do some sanity checks */
  if ((iNr < 0) || (iNr > 63))
    return (0);

  if (((iNr % 4) != 0) /* cannot delete packed AddIn */
   || ((model != 1) && (iNr == 0)))
        /* not X-model => don't delete build-in Add-In */
    return (0);

  p = MK_FP (0x0000, modeladdr[model]);
  if (!p)
    return (0);

  LinAddr = ((0x01100000 + (iNr << 14)) >> 14) & 0xFFFF;

  return (LinAddr != p[0]);
}


int DelAddIn (int iNr)
{
  unsigned int wSeg, wOffs;
  unsigned long LinAddr;
  int i;
  unsigned char far* pByte;
  char name[16];

  if (!CanDelAddIn (iNr))
    return (0);

      /* calculate start address of requested Add-In */
  LinAddr = 0x01100000 + (iNr << 14); /* Address of first Add-In for X-Models */

      /* make selected memory visible at 2000:0000 */
  SwitchBank ((((LinAddr) >> 16) & 0xFFFE), 1);

  wSeg = 0x2000;
  wOffs = LinAddr & 0xFFFF;
  if (LinAddr & 0x00010000)
    wSeg = 0x3000;

  if (deladdin (wSeg, wOffs))
  {
        /* Add-In is marked as deleted now, update Menu structure */
        /* You must call LibSaveSysRam() to make this permanent */

        /* search deleted Add-In in Menu structure */
    pByte = MK_FP (0x1840, 0x0B71);
    for (i = 0; i < 24; ++i)
      if (pByte[5 * i + 2] == IB_MADDIN)   /* Add-In found in, try to get it's name */
        if (LibGetProgramName (name, IB_MADDIN, pByte[5 * i + 1]) == FALSE)
          break;  /* this one is the deleted one */

    if (i < 24)
    {
      while (i < 23)
      {
            /* shift info for one Add-In */
        pByte[5 * i + 0] = pByte[5 * i + 5 + 0];
        pByte[5 * i + 1] = pByte[5 * i + 5 + 1];
        pByte[5 * i + 2] = pByte[5 * i + 5 + 2];
        pByte[5 * i + 3] = pByte[5 * i + 5 + 3];
        pByte[5 * i + 4] = pByte[5 * i + 5 + 4];
        ++i;
      }

          /* mark last Add-In as unused */
      pByte[5 * i + 0] = 0xFF;
      pByte[5 * i + 1] = 0xFF;
      pByte[5 * i + 2] = 0xFF;
      pByte[5 * i + 3] = 0xFF;
      pByte[5 * i + 4] = 0xFF;
    }

    return (1);
  }

  return (0);
}
#endif

static void near _read_icon (unsigned char* icon, unsigned long address,
                             unsigned int size, unsigned char xsize, unsigned char ysize)
{
  unsigned int wSeg, wOffs;
  char far* p;

        /* make memory for icon visible at 2000:0000 */
  SwitchBank (((address >> 16) & 0xFFFE), 1);

  wSeg = 0x2000;
  wOffs = address & 0xFFFF;
  if (address & 0x00010000)
    wSeg = 0x3000;

  p = MK_FP (wSeg, wOffs);

  far_memcpy (icon, p, size);
      /* correct dimensions if necessary */
  if (icon[0] > xsize)
    icon[0] = xsize;
  if (icon[2] > ysize)
    icon[2] = ysize;
  icon[1] = 0x00;
  icon[3] = 0x00;
  wOffs += size;
  if ((wOffs != 0) && (wOffs < size))
  {
        /* 64kB overflow */
    wSeg += 0x1000;
    if (wSeg == 0x4000)
    {
          /* correct bank */
      address += 0x4000;
      SwitchBank (((address >> 16) & 0xFFFE), 1);
      wSeg = 0x2000;
    }
    p = MK_FP (wSeg, 0);
    far_memcpy (&(icon[size - wOffs]), p, wOffs);
  }
}

int GetAddInInfo (int iNr, AddInInfo* pInfo, int ext)
{
  char far* p;
  unsigned long LinAddr, lTemp;
  unsigned int wSeg, wOffs;

      /* do some sanity checks */
  if ((iNr < 0) || (iNr > 63))
    return (0);

      /* calculate start address of requested Add-In */
  lTemp = iNr << 14;    /* each slot has a size of 16kB */
  LinAddr = 0x01100000  /* Address of first Add-In */ + (iNr << 14);

      /* make selected memory visible at 2000:0000 */
  SwitchBank (((LinAddr >> 16) & 0xFFFE), 1);

  wSeg = 0x2000;
  wOffs = LinAddr & 0xFFFF;
  if (LinAddr & 0x00010000)
    wSeg = 0x3000;

  p = MK_FP (wSeg, wOffs);

      /* test for Add-In header */
  if ((far_strncmp (&p[2], "CASIO", 5) == 0)
   && (far_strncmp (&p[8], "Z486", 4) == 0))
  {
#if FALSE
    if ((p[1] == 0) && !(iNr & 0x03)) /* if not packed and deleted */
      return (0);
#endif
    if (pInfo == NULL)
      return (1);

    pInfo->address = LinAddr;

      /* Add-In found, get header information */
    pInfo->flags = 0;
    if (p[1] == 0)
      pInfo->flags |= ADDININFO_DELETED;
    if (iNr & 0x03)
      pInfo->flags |= ADDININFO_PACKED;

    pInfo->mode    = p[19];
    pInfo->submode = p[18];

    far_strncpy (pInfo->name, &p[20], 16);
    pInfo->name[15] = 0;

    pInfo->size = ((unsigned long far*) p) [9];
    pInfo->orgsize = ((unsigned long far *)p)[19] + 84;
    if (pInfo->size != pInfo->orgsize)
      pInfo->flags |= ADDININFO_PACK;

    if (ext & ADDININFO_EXT_VERSION)
    {
        /* get version information */
      far_strncpy (pInfo->version, &p[40], 16);
      far_strncpy (pInfo->lib_version, &p[56], 16);

        /* get comment */
      far_strncpy (pInfo->comment, &p[0x50], MAXCOMMENTLENGTH);
      pInfo->comment [MAXCOMMENTLENGTH - 1] = 0;
      if (pInfo->comment[0] == 0xFF)
        pInfo->comment[0] = 0x00;
    }

    if (ext & ADDININFO_EXT_MENUICON)
    {   /* get menuicon */

        /* calculate address of menuicon */
      lTemp = ((unsigned long far*) p)[18] + LinAddr;

        /* read menuicon */
      _read_icon (pInfo->menuicon, lTemp, 176, 0x2D, 0x1C);
    }

    if (ext & ADDININFO_EXT_LISTICON)
    {   /* get listicon */

        /* calculate address of listicon */
      lTemp = ((unsigned long far*) p)[19] + LinAddr;

        /* read listicon */
      _read_icon (pInfo->listicon, lTemp, 84, 0x1B, 0x14);
    }

    return (1);
  }

  return (0);
}

#if FALSE
void CorrectMenuMaxPage (void)
{
  int n = 0;
  unsigned char far* p = MK_FP (0x1840,0x0B71);

  while (p[0] != 0xFF)
  {
    ++n;
    p += 5;
  }

  n = (n + 7) / 8;

  p = MK_FP (0x1840, 0x0B6E);
  p[1] = n;

  if (p[0] >= n)
    p[0] = n - 1;

      /* reset action menu pointer */
  p = MK_FP (0x1840, 0x00B0);
  *p = 0;
}
#endif

