
#include <stdrom.h>
#include "define.h"
#include "l_define.h"
#include "libc.h"
#include "l_libc.h"

#include "AddInInfo.h"
#include "modecall.h"



#define FLAGS_DELETEOLD 0x01
#define FLAGS_TEMPORARY 0x02
#define FLAGS_IMPORT    0x04


#ifdef LSI_C
 void _asm_breakpoint (char *);
 #define BreakPoint() _asm_breakpoint ("\n INT 03h\n")
#endif


const byte owbname[] = "OWBasic";

const byte FAR modulesig[] = {'O','B','I','M','P','O','R','T'};
#define OWBIMPORT_VER 0x00000003ul
#define CBSIGOFFS	2032

byte progname[32];

FILE_BUF  fb;
FILE_INFO finf;

byte flags;
byte startcat;


byte msg_exitsim[] = "READY TO EXIT SIM";


static const byte FAR kind[16] = {
  FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_TEXT,
  FILE_KIND_TEXT,FILE_KIND_BIN, FILE_KIND_BIN, FILE_KIND_TEXT,
  FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_BIN, FILE_KIND_BIN,
  FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_TEXT,FILE_KIND_TEXT};

/* transform OWB submode -> PVOS submode + secretinfo */
#define Sub(sub)  (sub & 0x0F)
#define Info(sub) (((sub) > 15) ? FILE_SECRET_MODE : FILE_OPEN_MODE)

static void NEAR SetFP (int mode, int sub, word fp)
{
  fb.fsb_main_entry_ = mode << 4;	/* Main */
  fb.fsb_sub_entry_  = Sub (sub);	/* Sub */
  fb.fsb_scrt_info_  = Info (sub);	/* Open Mode */

  fb.fsb_todo_chek_ = 0xFF;
  fb.fsb_todo_rank_ = 0xFF;

  finf.fp      = fp;
  finf.kind    = kind[mode];
}


static int NEAR namedfile (const uchar* name, word* imode, word* isubmode)
{
  byte submode, mode;

  while (!LibGetAllEntry ((byte*) name, &mode, &submode)) /* get mode and submode */
    if (!LibSubEntrySave ((byte*) name, &submode)) /* if not exists, create new one */
      return (1);

  *imode = mode >> 4; /* convert to OWB style */
  *isubmode = submode;
  return (0);
}


static int NEAR filefind (int mode, int submode, const uchar* name, const uchar* sig, word* fp)
{
  uchar sname[220]; /* name to search for */

  bool found = FALSE;
  uchar* ip = sname;
  byte* bp;
  /*int i;*/
  register uchar bc, ic;
  bool is_bin = (kind[mode] == FILE_KIND_BIN);


  while (*(ip++) = tolower (*(name++)))
    ;
  *ip = 0x00;

  finf.fp = 0xFFFF; /* set file pointer for first search */
  do
  {
    SetFP (mode, submode, finf.fp);
    LibFileFindNext (&fb, &finf, 0x00);

    if (LibFileReadEx (&fb, &finf , (220 + FILE_BLOCK_SIZE - 1) / FILE_BLOCK_SIZE))
    {
      ip = sname - 1;
      bp = is_bin ? fb.fbuf.bin.bin_buf : fb.fbuf.text.fsb_text_buf_;
      found = TRUE;

      if (sig)
        if (*sig)
        {
          if ((*bp == *sig) && (bp[1] == sig[1]))
          {
            bp += 2;
            while (*bp == ' ')
              ++bp; /* skip whitespaces after signature */
          }
          else
            found = FALSE;
        }

      --bp;

      if (found)
      {
        found = FALSE;
        do
        {
          bc = *(++bp);
          ic = *(++ip);
          if (((bc == 0x0A) || (bc == 0x0D) || (bc == 0x00)) && (ic == 0x00))
          {
            *fp = finf.fp;
            return (TRUE);
          }
          if (ic != tolower (bc))
            break;
        }
        while (ip < (sname + 220 - 1));
      }
    }
  } while (!found && (finf.fp != 0xFFFF));

  return (found);
}


static void NEAR writeFiles (void)
{
  word filecnt, cat;
  dword fileoffs[64];
  AddInInfo aii;

  byte buf[220];
  int n;
  int tfp;
  word seg, offs;
  byte  FAR* pb;
  word  FAR* pw;
  dword FAR* pd;
  bool unchanged;

  byte FAR* sp; byte* dp;
  

    /* find appended memory */
  for (n = 0; n < 64; ++n)
    if (GetAddInInfo (n, &aii, ADDININFO_EXT_NONE))
      if (strcmp (aii.name, owbname)== 0)
        break;
  if (n == 64)
  {
    LibPutMsgDlg ("OWBasic not found");
    LibJumpMenu ();
  }
  

    /* memory found, make visible at bank 5 (A000:0000) */
  SwitchBank (((aii.address + (8 /* 8 16kB segments */ << 14)) >> 16) & 0xFFFE, 5);


    /* validate signature and version */
  pb = MK_FP (0xA000, 0x0000);
  if (far_memcmp (pb, modulesig, 8))
  {
    LibPutMsgDlg ("Wrong data format");
    LibJumpMenu ();
  }

  pd = MK_FP (0xA000, 0x0008);
  if (*pd != OWBIMPORT_VER)
  {
	LibPutMsgDlg ("Wrong data version");
    LibJumpMenu ();
  }


    /* readout basic information */
  pw = MK_FP (0xA000, 0x000C);
  filecnt  = *pw;
  startcat = pb[14];
  flags    = pb[15];
  
  far_strcpy (progname, pb + 16);
  far_memcpy (fileoffs, pb + 48, 64 * sizeof (dword));
  

    /* write files */
  for (n = 0; n < filecnt; ++n)
  {
    seg = ((0xA000U >> 12) + (fileoffs[n] >> 16)) << 12;
	offs = fileoffs[n] & 0xFFFF;
	pb = MK_FP (seg, offs);
	
	cat       = *(pb++);
	unchanged = *(pb++);
    
	
	  /* copy filename to buf */
	dp = buf - 1; sp = pb;
	do
	  *++dp = *(sp++);
	while ((*dp != 0x00) && (*dp != 0x0A) && (*dp != 0x0D));
	*dp = 0x00;

	LibClrDisp ();
	LibPutProStr (IB_PFONT3, 1, 1, "OBIMPORT", 159);
	LibPutProStr (IB_PFONT1, 1, 15, "Transferring files", 159);
	LibPutProStr (IB_PFONT1, 1, 25, buf, 159);
	  
	SetFP (0x03, cat, 0xFFFF);
	if (flags & FLAGS_DELETEOLD)
	{
	  if (!filefind (0x03, cat, buf, NULL, &tfp))
		tfp = 0xFFFF;
      else if (unchanged)
          continue;
	  SetFP (0x03, cat, tfp);
	}
    else if (unchanged)
        continue;
	LibPutDisp ();
	
	far_strcpy (fb.fbuf.text.fsb_text_buf_, pb);

    if (!LibFileWrite (&fb, &finf))
    {
	  LibPutMsgDlg ("Can't write file!\r\rPerform Memory\r\rManagement!");
      LibJumpMenu ();
    }
  }
}


static void NEAR startOWB (void)
{
  word mode;

  mode = mc_FindMode (owbname);
  mc_InitWrite ();
  mc_WriteParam (startcat, progname, 0xFFFF);
  mc_WriteEmpty ();
  mc_WriteEmpty ();
  LibModeJump (mode, 0x00E1);
}


static void NEAR writeReturnInfo (void)
{
  word fp;
  word mode, submode;
  byte FAR* clipbrd = SYS_INP_BUF;
  byte* filebuf = fb.fbuf.bin.bin_buf;
  
  namedfile ("SHARED", &mode, &submode);

  if (!filefind (mode, submode, "OWBasic,rtmsg", NULL, &fp))
  {
	LibPutMsgDlg ("Can't find error message");
	return;
  }

  while (*(filebuf++) != 0x00) /* skip title */
	;
  while (*(clipbrd++) != ':')  /* skip file and line */
	;
  ++clipbrd; /* skip white space after ':' */
	/* replace "Error position found" with runtime error message */
  far_strcpy (clipbrd, filebuf);
  far_strcat (clipbrd, "(Runtime error)");
}


void emptyClipboardSig (void)
{
  int n;

  SYS_INP_BUF[0] = '\0';
  for (n = 0; n < 12; ++n)
	SYS_INP_BUF[CBSIGOFFS + n] = '\0';
}
void writeClipboardSig (void)
{
  far_memcpy (SYS_INP_BUF + CBSIGOFFS, modulesig, 8);
  *(long FAR*) (SYS_INP_BUF + CBSIGOFFS + 8) = OWBIMPORT_VER;
}

void main (void)
{
  word seg, offs;
  word mode, status;

  uchar namebuf[50];
  int cat, curpos;
  
  TCHSTS tsts;
  
  
  LibTchStackClr ();
  LibTchStackPush (TchAllDsp);
  LibTchStackPush (TchHardIcon);

  LibGetMode (&mode, &status, &seg, &offs);

  emptyClipboardSig ();
  
  if (status == 0x8000) /* DualWin mode - started as Tools */
  { /* Tools are not available */
    LibPutMsgDlg ("Tools not available");
    LibDualWinExit (NULL);
  }

  if (status == 0x4000) /* started by initialisation */
  { /* read out file properties and start OWBasic */
	writeFiles ();
    startOWB ();
  }
  
  if ((status & 0xFFF0) == 0x00F0) /* return from OWBasic */
  {
	mc_ReadParam (&cat, namebuf, &curpos);
	if ((curpos != 0) || (status & 0x0001) /* || SYS_INP_BUF[0]*/) /* error */
	{
		if (status == 0x00F1) /* runtime error */
			writeReturnInfo ();
		writeClipboardSig ();
	}
	else
		emptyClipboardSig ();
	
    LibClrDisp ();
	LibPutProStr (IB_PFONT3, 21, 73, msg_exitsim, 159);
    LibPutDisp ();
    LibTchWait (&tsts);
    LibJumpMenu ();
  }

  LibPutMsgDlg ("Unknown status code:\r0x%X", status);
  LibJumpMenu ();
}




