//---------------------------------------------------------------------------

#include <cstring>
#include <string>
#include <utility>
#include <vector>
#include <vcl.h>
#include <windows.h>
#include "Twfdec.hpp"
#include "ucl/file.hpp"
#pragma hdrstop
#include "twf2obp_psel.h"
//---------------------------------------------------------------------------
USEFORM("twf2obp_u1.cpp", FrmMain);
USEFORM("twf2obp_psel.cpp", FrmSelProgram);
//---------------------------------------------------------------------------
const std::vector <std::pair <AnsiString, int> >* globalProgramNames = NULL;
int globalProgramIndex = -1;

bool cmdLineStart (void)
{
    int cnt = ParamCount ();
    
    if (cnt < 2)
        return (false);

    if (ParamStr (1) != "-r")
        return (false);

    return (true);
}


static __stdcall void receiveMessage (char* text, int i)
{
    if (std::strncmp (text, "Error", 5) == 0)
        *reinterpret_cast <bool*> (i) = false;
}


std::string intToString (int i, int radix = 10)
{
	char buf[33+1];

	return (itoa (i, buf, radix));
}

bool createProjectFile (const char* path, AnsiString& projfile, bool cmdline)
{
    ucl::IniFile ofile;
    std::vector <AnsiString> files;
    std::vector <std::pair <AnsiString, int> > prognames;
    WIN32_FIND_DATA wfd;
    HANDLE hSrch;
    AnsiString str;


    globalProgramIndex = -1;
    globalProgramNames = &prognames;
	ofile["Settings"]["ProjectFormatVersion"] = intToString (120);

        /* enumerate directories and files */
    for (int i = 1; i <= 5; ++i)
        if (DirectoryExists (path + AnsiString (i)))
        {
            hSrch = FindFirstFile ((path + AnsiString (i) + "\\*.*").c_str (), &wfd);
            if (!hSrch)
                continue;
            do
            {
                if ((std::strcmp (wfd.cFileName, ".") == 0)
                  || std::strcmp (wfd.cFileName, "..") == 0)
                    continue;
                str = path + AnsiString (i) + '\\' + wfd.cFileName;
                files.push_back (str);
                ofile["Category " + intToString (i)].addLine (intToString (i) + "\\" + ExtractFileName (str).c_str ());
                if (ExtractFileExt (wfd.cFileName) == ".bas")
                    prognames.push_back (std::pair <AnsiString, int> (
                        ExtractFileName (str.SubString (1, str.Length () - 4)), i));
            } while (FindNextFile (hSrch, &wfd));
            FindClose (hSrch);
        }
    
    if (prognames.size () > 1)
    {
        if (cmdline)
        {
            Application->CreateForm (__classid(TFrmSelProgram), &FrmSelProgram);
            Application->Run ();
        }
        else
            FrmSelProgram->ShowModal ();
        if (globalProgramIndex < 0)
            return (false);
    }
    else
        globalProgramIndex = 0;

    projfile = prognames[globalProgramIndex].first;

    ofile["Settings"]["Progname"] = projfile.c_str ();
    ofile["Settings"]["Category"] = intToString (prognames[globalProgramIndex].second);

    projfile = path + projfile + ".obp";
    ofile.saveToFile (projfile.c_str ());
    
    return (prognames.size () != 0);
}

AnsiString getExePath (void)
{
	char exepath[1024];
	int i;

	GetModuleFileName (NULL, exepath, 1024);
	
	return (ExtractFilePath (exepath));
}

void startInSimulator (const AnsiString& projfile)
{  /* start with owb2sim */
	STARTUPINFO si;
    PROCESS_INFORMATION pi;
    AnsiString cmdline = getExePath () + "owb2sim.exe -e \"" + projfile + '"';


	si.cb = sizeof (STARTUPINFO);
	si.lpReserved = NULL;
	si.lpDesktop = NULL;
	si.lpTitle = NULL;
	si.cbReserved2 = 0;
	si.lpReserved2 = NULL;
	si.dwFlags = 0;
	CreateProcess (NULL, &cmdline[1], NULL, NULL, FALSE,
		NORMAL_PRIORITY_CLASS | CREATE_NO_WINDOW, NULL, ExtractFilePath (projfile).c_str (), &si, &pi);
    CloseHandle (pi.hProcess);
    CloseHandle (pi.hThread);
}

void startTWFProject (void)
{
    char temp[MAX_PATH + 5];
    SHFILEOPSTRUCT fos;
    bool success = true;
    AnsiString projfile;


    GetTempPath (MAX_PATH, temp);
    std::strcat (temp, "~twf2obp0\\");

    CreateDirectory (temp, NULL); /* create if necessary */
    std::strcat (temp, "*.*");
    temp[std::strlen (temp) + 1] = '\0';

    ZeroMemory (&fos, sizeof (SHFILEOPSTRUCT));
    fos.wFunc = FO_DELETE;
    fos.pFrom = temp;
    fos.fFlags = FOF_NOCONFIRMATION | FOF_SILENT;
    SHFileOperation (&fos); /* remove directory contents */
    temp[std::strlen (temp) - 3] = '\0'; /* remove "*.*" */

    decodeTWFFile (temp, ParamStr (2).c_str (),
                   twfMemo, decTXT | decSingleMemo | decOWBExtensions,
                   receiveMessage, reinterpret_cast <int> (&success));

    if (!success)
        return;

    if (!createProjectFile (temp, projfile, true))
        return;

    startInSimulator (projfile);
}


WINAPI WinMain(HINSTANCE, HINSTANCE, LPSTR, int)
{
    /*
     *  command line parameters:
     *  twf2obp [-r <twfpath>]
     *
     *   -r <twfpath> - extract files from TWF and start program
     *
     *
     */

    try
    {
        Application->Initialize ();

        if (cmdLineStart ())
        {   /* extract project to temporary directory and start it */
            startTWFProject ();
        }
        else  /* create FrmMain */
        {
            Application->CreateForm(__classid(TFrmMain), &FrmMain);
            Application->CreateForm (__classid(TFrmSelProgram), &FrmSelProgram);
            Application->Run();
        }
    }
    catch (Exception &exception)
    {
         Application->ShowException(&exception);
    }
    catch (...)
    {
         try
         {
             throw Exception("");
         }
         catch (Exception &exception)
         {
             Application->ShowException(&exception);
         }
    }
    return 0;
}
//---------------------------------------------------------------------------
