

#include "file.hpp"
#include <cstdio>


namespace ucl
{

std::string extractFileName (const std::string& path)
{
	int slast = path.size () - 1;
	int nfirst, nlast = slast;
	
	for (nfirst = slast; (path[nfirst] != '\\') && (nfirst > 0); --nfirst)
	{
		if ((nlast == slast) && (path[nfirst] == '.'))
			nlast = nfirst - 1;
	}
	if (path[nfirst] == '\\')
		++nfirst;
	
	return (path.substr (nfirst, nlast + 1 - nfirst));
}
std::string extractFilePath (const std::string& path)
{
	int i;

	for (i = path.size () - 1; (path[i] != '\\') && (i > 0); --i)
		;
	if (path[i] == '\\')
		++i;
	
	return (path.substr (0, i));
}
std::string extractFileExt  (const std::string& path)
{
	int i;

	for (i = path.size () - 1; (path[i] != '.') && (i > 0); --i)
		if (path[i] == '\\')
			break;;
	if (path[i] != '.')
		return ("");
	
	return (path.substr (i));
}




BasicTextFile::stringlist_t::iterator BasicTextFile::getIteratorFromIndex (stringlist_t& sl, unsigned int idx)
{
	stringlist_t::iterator it = sl.begin ();
    if (idx < sl.size ())
        while (idx > 0)
            ++it, --idx;
    return (it);
}
bool BasicTextFile::read_string (std::FILE* stream, std::string& str)
{
    char c;

    str.clear ();
    while (true)
    {
        c = fgetc (stream);
        if (c == '\n')
            break;
        else if (c == EOF)
            return (true);
        str += c;
    }
    return (false);
}


bool TextFile::load_internal (const std::string& name) const
{
    std::FILE* file;
    std::string buf;
    char c;
    bool eof;

    buf.reserve (100);
    lines.clear ();

	file = std::fopen (name.c_str (), "r");
	if (!file)
		return (false);
	std::fseek (file, 0, SEEK_SET);

    do
    {
        eof = read_string (file, buf);
        lines.push_back (buf);
    } while (!eof);

    if (lines.back ().empty ())
        lines.pop_back (); /* remove last line, of empty */

	std::fclose (file);
	return (true);
}

bool TextFile::saveToFile (const std::string& name) const
{
	std::FILE* file;
	stringlist_t::iterator it = lines.begin ();
	stringlist_t::iterator end = lines.end ();

	file = std::fopen (name.c_str (), "w");
	if (!file)
		return (false);
	std::fseek (file, 0, SEEK_SET);

	while (it < end)
		std::fputs (((*(it++)) + '\n').c_str (), file);

	std::fclose (file);
	return (true);
}

bool TextFile::loadFromFile (const std::string& name)
    { return (load_internal (name)); }
void TextFile::clear (void)
	{ lines.clear (); }



std::string IniFile::extractText (const std::string& str)
{
	int slast = str.size () - 1;
	int nfirst;

	for (nfirst = slast; (str[nfirst] != '=') && (nfirst > 0); --nfirst)
		;
	if (str[nfirst] == '=')
		++nfirst;

	return (str.substr (nfirst));
}
std::string IniFile::extractVar (const std::string& str)
{
	int slast = str.size ();
	int eq;

	for (eq = 0; eq < slast; ++eq)
		if (str[eq] == '=')
			return (str.substr (0, eq));
	return (std::string ());
}

IniFile::StringProxy IniFile::Section::operator [] (const std::string& var)
{
	stringlist_t::iterator it = lines.begin ();
	stringlist_t::iterator end = lines.end ();

	while (it < end)
	{
		if (extractVar (*it) == var)
			return (StringProxy (*it));
		++it;
	}
		/* if not available, create new entry */
	lines.push_back (var + '=');
	return (StringProxy (lines.back ()));
}

const IniFile::StringProxy IniFile::Section::operator [] (const std::string& var) const
{
	stringlist_t::iterator it = lines.begin ();
	stringlist_t::iterator end = lines.end ();

	while (it < end)
	{
		if (extractVar (*it) == var)
			return (StringProxy (*it));
		++it;
	}
		/* if not available, throw exception */
	throw ("E_NOT_FOUND");
}

bool IniFile::Section::entryExists (const std::string& entry) const
{
	stringlist_t::iterator it = lines.begin ();
	stringlist_t::iterator end = lines.end ();

	while (it < end)
	{
		if (extractVar (*it) == entry)
			return (true);
		++it;
	}
	return (false);
}



IniFile::Section& IniFile::operator [] (const std::string& section)
{
	std::vector <Section>::iterator it = sections.begin ();
	std::vector <Section>::iterator end = sections.end ();

	while (it < end)
	{
		if (it->name == section)
			return (*it);
		++it;
	}
		/* if not available, create new entry */
	sections.push_back (Section (section));
    return (sections.back ());
}
const IniFile::Section& IniFile::operator [] (const std::string& section) const
{
	std::vector <Section>::const_iterator it = sections.begin ();
	std::vector <Section>::const_iterator end = sections.end ();

	while (it < end)
	{
		if (it->name == section)
			return (*it);
		++it;
	}
		/* if not available, throw exception */
	throw ("E_NOT_FOUND");
}

void IniFile::addSection (const std::string& section)
{ sections.push_back (Section (section)); }

bool IniFile::sectionExists (const std::string& section) const
{
	std::vector <Section>::const_iterator it = sections.begin ();
	std::vector <Section>::const_iterator end = sections.end ();

	while (it < end)
	{
		if (it->name == section)
			return (true);
		++it;
	}
	return (false);
}


bool IniFile::load_internal (const std::string& name) const
{
    std::FILE* file;
    std::string buf;
    bool eof;
    int l, i;

    buf.reserve (100);
    sections.clear ();

    file = std::fopen (name.c_str (), "r");
    if (!file)
        return (false);
    std::fseek (file, 0, SEEK_SET);

    do
    {
		eof = read_string (file, buf);
        l = buf.size () - 1;
        i = 0;
        if (l > 0)
        {
          while   ((i <= l) && buf[i] == ' ')
                ++i; /* skip leading space */
            while (l >= 0 && buf[l] == ' ')
                --l; /* skip trailing space */

          /*if (buf[i] == ';') /+ skip comment +/
              continue;*/
            if ((buf[i] == '[') && (buf[l] == ']')) /* section header */
                sections.push_back (Section (buf.substr (i + 1, l - i - 1)));
			else if (sections.size () > 0)
				sections.back ().addLine (buf.substr (i));
		}

    } while (!eof);

	std::fclose (file);
	return (true);
}
bool IniFile::saveToFile (const std::string& name) const
{
	std::FILE* file;
	std::vector <Section>::const_iterator it = sections.begin ();
	std::vector <Section>::const_iterator end = sections.end ();

	int i, l;

	file = std::fopen (name.c_str (), "w");
	if (!file)
		return (false);
	std::fseek (file, 0, SEEK_SET);

	while (it < end)
	{
		std::fputs (('[' + it->name + "]\n").c_str (), file);

		l = it->lineCount ();
		for (i = 0; i < l; ++i)
			std::fputs (((*it)[i] + '\n').c_str (), file);

		std::fputs ("\n", file);

		++it;
	}

	std::fclose (file);
	return (true);
}

void IniFile::clear (void)
{ sections.clear (); }
std::string& TextFile::operator [] (unsigned int idx)
{ return (lines.at (idx)); }
const std::string& TextFile::operator [] (unsigned int idx) const
{ return (lines.at (idx)); }
void TextFile::addLine (const std::string& text)
{ lines.push_back (text); }
void TextFile::insertLine (unsigned int line, const std::string& text)
{ lines.insert (getIteratorFromIndex (lines, line), text); }
void TextFile::deleteLine (unsigned int line)
{ lines.erase (getIteratorFromIndex (lines, line)); }


} /* end namespace ucl */
